'use strict';

var Curry = require("bs-platform/lib/js/curry.js");
var Belt_List = require("bs-platform/lib/js/belt_List.js");
var Belt_Option = require("bs-platform/lib/js/belt_Option.js");
var Caml_option = require("bs-platform/lib/js/caml_option.js");
var NormalizrNew = require("./NormalizrNew.js");
var Caml_exceptions = require("bs-platform/lib/js/caml_exceptions.js");
var Caml_builtin_exceptions = require("bs-platform/lib/js/caml_builtin_exceptions.js");

function objNullableToRecordOptional(obj, convertToRecord) {
  if (obj == null) {
    return ;
  } else {
    return Caml_option.some(Curry._1(convertToRecord, obj));
  }
}

function recordOptionalToObjNullable(record, convertToObj) {
  if (record !== undefined) {
    return Curry._1(convertToObj, Caml_option.valFromOption(record));
  } else {
    return null;
  }
}

function getConnectionList(items, idFunction) {
  var __x = Belt_List.fromArray(items.edges);
  return Belt_List.map(__x, (function (edge) {
                return Belt_Option.map(edge, (function (e) {
                              return Curry._1(idFunction, e.node);
                            }));
              }));
}

function GenerateModel(Root, $star) {
  var Schema = Caml_exceptions.create("ModelUtils.GenerateModel(Root)(*).Schema");
  var Id = Caml_exceptions.create("ModelUtils.GenerateModel(Root)(*).Id");
  var idToString = function (id) {
    var match = id[0];
    if (match[0] === Id) {
      return match[1];
    } else {
      return "";
    }
  };
  var idToRootId = function (id) {
    var match = id[0];
    if (match[0] === Id) {
      return [
              Id,
              match[1]
            ];
    } else {
      throw [
            Caml_builtin_exceptions.match_failure,
            /* tuple */[
              "ModelUtils.re",
              42,
              4
            ]
          ];
    }
  };
  var stringToId = function (uuid) {
    return /* tuple */[
            [
              Id,
              uuid
            ],
            /* Type */0
          ];
  };
  return {
          Root: Root,
          Id: Id,
          Schema: Schema,
          idToString: idToString,
          idToRootId: idToRootId,
          stringToId: stringToId
        };
}

function AddModel(SchemaType, ModelType) {
  var Record = Caml_exceptions.create("ModelUtils.AddModel(SchemaType)(ModelType).Record");
  var wrap = function (model) {
    return [
            Record,
            model
          ];
  };
  var unwrap = function (record) {
    if (record[0] === Record) {
      return record[1];
    }
    
  };
  return {
          Model: ModelType,
          Wrapper: {
            wrap: wrap,
            unwrap: unwrap,
            apolloEnabled: true
          },
          Record: Record
        };
}

function AddRecord(Record, NormalizrGenerator) {
  var modelIdToIdFunction = function (id) {
    if (id[0] === Record.Model.ModelSchemaType.Id) {
      return /* tuple */[
              Record.Model.ModelSchemaType.Schema,
              id[1]
            ];
    } else {
      return Curry._1(NormalizrGenerator.modelIdToIdFunction, id);
    }
  };
  var modelTypeToRecordType = function (record) {
    if (record[0] === Record.Record) {
      return /* tuple */[
              Record.Model.ModelSchemaType.Schema,
              Curry._1(Record.Model.Record.findId, record[1])
            ];
    } else {
      return Curry._1(NormalizrGenerator.modelTypeToRecordType, record);
    }
  };
  var normalizerCommitItemToSchema = function (param, param$1) {
    return NormalizrNew.Normalizr.commitItemToSchema(modelTypeToRecordType, param, param$1);
  };
  return {
          modelIdToIdFunction: modelIdToIdFunction,
          modelTypeToRecordType: modelTypeToRecordType,
          normalizerCommitItemToSchema: normalizerCommitItemToSchema
        };
}

function EmptyNormalizr(Root) {
  var modelIdToIdFunction = function (id) {
    id === Root.EMPTY_ID;
    return /* tuple */[
            Root.EMPTY_T,
            ""
          ];
  };
  var modelTypeToRecordType = function (record) {
    record === Root.EMPTY_RECORD;
    return /* tuple */[
            Root.EMPTY_T,
            ""
          ];
  };
  var normalizerCommitItemToSchema = function (param, param$1) {
    return NormalizrNew.Normalizr.commitItemToSchema(modelTypeToRecordType, param, param$1);
  };
  return {
          modelIdToIdFunction: modelIdToIdFunction,
          modelTypeToRecordType: modelTypeToRecordType,
          normalizerCommitItemToSchema: normalizerCommitItemToSchema
        };
}

function CreateFakeLocal($star) {
  var _defaultRecord = function (id) {
    return /* () */0;
  };
  var $$default = function (id) {
    return /* () */0;
  };
  var Record = {
    $$default: $$default
  };
  return {
          _defaultRecord: _defaultRecord,
          Record: Record
        };
}

function BuildModel(ModelRecord, ModelSchemaType, Fragment) {
  var getUUIDFromId = function (id) {
    return Curry._1(ModelSchemaType.idToString, id);
  };
  var idToTypedId = function (id) {
    return Curry._1(ModelSchemaType.stringToId, id);
  };
  var fragmentName = Fragment.Fields.name;
  var objectToId = function (obj) {
    return Curry._1(ModelSchemaType.stringToId, Curry._1(Fragment.toId, obj));
  };
  var Data = {
    fromObject: Fragment.fromObject
  };
  var defaultWithId = function (param, id) {
    return Curry._2(ModelRecord._defaultWithId, param, Curry._1(ModelSchemaType.idToString, id));
  };
  var fromObject = function (obj) {
    return {
            data: Curry._1(Fragment.fromObject, obj),
            local: Curry._1(ModelRecord.Local.Record.$$default, Curry._1(Fragment.toId, obj))
          };
  };
  return {
          ModelSchemaType: ModelSchemaType,
          InternalSchema: ModelSchemaType.Root,
          ModelRecord: ModelRecord,
          idToRootId: ModelSchemaType.idToRootId,
          getUUIDFromId: getUUIDFromId,
          idToTypedId: idToTypedId,
          Fragment: Fragment,
          Record: {
            findId: ModelRecord.findId,
            $$default: ModelRecord._defaultRecord,
            defaultParamData: ModelRecord.defaultParamData,
            Data: Data,
            fromObject: fromObject,
            defaultWithId: defaultWithId
          },
          objectToId: objectToId,
          fragmentType: Fragment.fragmentType,
          fragmentName: fragmentName,
          _defaultData: ModelRecord._defaultData
        };
}

exports.objNullableToRecordOptional = objNullableToRecordOptional;
exports.recordOptionalToObjNullable = recordOptionalToObjNullable;
exports.getConnectionList = getConnectionList;
exports.GenerateModel = GenerateModel;
exports.AddModel = AddModel;
exports.AddRecord = AddRecord;
exports.EmptyNormalizr = EmptyNormalizr;
exports.CreateFakeLocal = CreateFakeLocal;
exports.BuildModel = BuildModel;
/* No side effect */
