'use strict';

var Curry = require("bs-platform/lib/js/curry.js");
var Caml_obj = require("bs-platform/lib/js/caml_obj.js");
var Belt_List = require("bs-platform/lib/js/belt_List.js");
var Pervasives = require("bs-platform/lib/js/pervasives.js");
var Belt_Option = require("bs-platform/lib/js/belt_Option.js");

function addOrModifyById(recordList, rType, newUUID) {
  var maybeRecordTuple = Belt_List.getBy(recordList, (function (param) {
          return Caml_obj.caml_equal(param[0], newUUID);
        }));
  if (maybeRecordTuple !== undefined) {
    return Belt_List.map(recordList, (function (param) {
                  var iterUUID = param[0];
                  var match = Caml_obj.caml_equal(newUUID, iterUUID);
                  if (match) {
                    return /* tuple */[
                            newUUID,
                            rType
                          ];
                  } else {
                    return /* tuple */[
                            iterUUID,
                            param[1]
                          ];
                  }
                }));
  } else {
    return Pervasives.$at(recordList, /* :: */[
                /* tuple */[
                  newUUID,
                  rType
                ],
                /* [] */0
              ]);
  }
}

function commitItemToSchema(typeConversionFunction, normalizedSchema, rType) {
  var match = Curry._1(typeConversionFunction, rType);
  var id = match[1];
  var newSchmaType = match[0];
  var maybeSchmaType = Belt_List.getBy(normalizedSchema, (function (param) {
          return Caml_obj.caml_equal(param[0], newSchmaType);
        }));
  if (maybeSchmaType !== undefined) {
    return Belt_List.map(normalizedSchema, (function (param) {
                  var iterRecordList = param[1];
                  var iterSchemaType = param[0];
                  var match = Caml_obj.caml_equal(iterSchemaType, newSchmaType);
                  if (match) {
                    return /* tuple */[
                            iterSchemaType,
                            addOrModifyById(iterRecordList, rType, id)
                          ];
                  } else {
                    return /* tuple */[
                            iterSchemaType,
                            iterRecordList
                          ];
                  }
                }));
  } else {
    return Pervasives.$at(normalizedSchema, /* :: */[
                /* tuple */[
                  newSchmaType,
                  addOrModifyById(/* [] */0, rType, id)
                ],
                /* [] */0
              ]);
  }
}

function getItemFromSchema(modelIdToIdFunction, normalizedSchema, modelIdType) {
  var match = Curry._1(modelIdToIdFunction, modelIdType);
  var id = match[1];
  var schemaType = match[0];
  var maybeRecordListTuple = Belt_List.getBy(normalizedSchema, (function (param) {
          return Caml_obj.caml_equal(schemaType, param[0]);
        }));
  if (maybeRecordListTuple !== undefined) {
    var __x = Belt_List.getBy(maybeRecordListTuple[1], (function (param) {
            return Caml_obj.caml_equal(param[0], id);
          }));
    return Belt_Option.map(__x, (function (param) {
                  return param[1];
                }));
  }
  
}

var Normalizr = {
  defaultNormalized: /* [] */0,
  addOrModifyById: addOrModifyById,
  commitItemToSchema: commitItemToSchema,
  getItemFromSchema: getItemFromSchema
};

exports.Normalizr = Normalizr;
/* No side effect */
