'use strict';

var Curry = require("bs-platform/lib/js/curry.js");
var Belt_List = require("bs-platform/lib/js/belt_List.js");
var Pervasives = require("bs-platform/lib/js/pervasives.js");
var Belt_Option = require("bs-platform/lib/js/belt_Option.js");
var Caml_option = require("bs-platform/lib/js/caml_option.js");
var NormalizrNew = require("./NormalizrNew.js");

function tryAgainifNullOption(optionA, optionB) {
  if (optionA !== undefined) {
    return optionA;
  } else {
    return optionB;
  }
}

function removeOptionsFromList(list) {
  return Belt_List.reduce(list, /* [] */0, (function (memo, ele) {
                if (ele !== undefined) {
                  return Pervasives.$at(memo, /* :: */[
                              Caml_option.valFromOption(ele),
                              /* [] */0
                            ]);
                } else {
                  return memo;
                }
              }));
}

function DomainTypeConverter(NormalizrGenerator, DomainType) {
  var Container = DomainType.Container;
  var Wrapper = DomainType.Record.Wrapper;
  var normalizerGetItemFromSchema = function (param, param$1) {
    return NormalizrNew.Normalizr.getItemFromSchema((function (a) {
                  return Curry._1(NormalizrGenerator.modelIdToIdFunction, Curry._1(DomainType.Model.idToRootId, a));
                }), param, param$1);
  };
  var idListToFilteredItems = function (idList, modelTypefunction) {
    return removeOptionsFromList(Belt_List.map(idList, modelTypefunction));
  };
  var reduce = function ($$default, getterFn, reduce$1, normalized, id, action) {
    var modelPromise = Curry._2(reduce$1, action, normalized.then((function (norm) {
                var __x = Curry._2(getterFn, norm, id);
                return Promise.resolve(Belt_Option.getWithDefault(__x, Curry._1($$default, id)));
              })));
    return Promise.all(/* tuple */[
                  normalized,
                  modelPromise
                ]).then((function (param) {
                  return Promise.resolve(Curry._2(NormalizrGenerator.normalizerCommitItemToSchema, param[0], Curry._1(Wrapper.wrap, param[1])));
                }));
  };
  var fromLocal = function (normalized, id) {
    var __x = normalizerGetItemFromSchema(normalized, id);
    return Belt_Option.flatMap(__x, Wrapper.unwrap);
  };
  var fromSchema = function (normalized, id) {
    var optionNormalized = fromLocal(normalized, id);
    if (Wrapper.apolloEnabled) {
      var __x = Curry._1(Container.getById, id);
      return tryAgainifNullOption(Belt_Option.map(__x, (function (fragment) {
                        if (optionNormalized !== undefined) {
                          return {
                                  data: Curry._1(DomainType.Model.Record.Data.fromObject, fragment),
                                  local: optionNormalized.local
                                };
                        } else {
                          return Curry._1(DomainType.Model.Record.fromObject, fragment);
                        }
                      })), optionNormalized);
    } else {
      return optionNormalized;
    }
  };
  var GenerateConverterFunc = function (SourceContainer) {
    var getRecord = SourceContainer.getRecord;
    var getRecordWithDefault = function (normalized, id, param) {
      var __x = Curry._2(SourceContainer.getRecord, normalized, id);
      return Belt_Option.getWithDefault(__x, Curry._2(DomainType.Model.Record.defaultWithId, param, id));
    };
    var idListToFilteredItems = function (idList, normalized) {
      return removeOptionsFromList(Belt_List.map(idList, Curry._1(SourceContainer.getRecord, normalized)));
    };
    var reduceWithDefault = function (param) {
      return (function (param$1, param$2, param$3) {
          return reduce((function (__x) {
                        return Curry._2(DomainType.Model.Record.defaultWithId, param, __x);
                      }), getRecord, DomainType.Action.reduce, param$1, param$2, param$3);
        });
    };
    var createReduceIdWithDefault = function (normalized, id, param, action) {
      return reduceWithDefault(param)(Promise.resolve(normalized), id, action);
    };
    var createPromiseReduceIdWithDefault = function (normalizedPromise, id, param, action) {
      return reduceWithDefault(param)(normalizedPromise, id, action);
    };
    return {
            getRecord: getRecord,
            getRecordWithDefault: getRecordWithDefault,
            idListToFilteredItems: idListToFilteredItems,
            reduceWithDefault: reduceWithDefault,
            createReduceIdWithDefault: createReduceIdWithDefault,
            createPromiseReduceIdWithDefault: createPromiseReduceIdWithDefault
          };
  };
  var LocalSourceContainer = {
    getRecord: fromLocal
  };
  var RemoteSourceContainer = {
    getRecord: fromSchema
  };
  var getRecordWithDefault = function (normalized, id, param) {
    var __x = fromLocal(normalized, id);
    return Belt_Option.getWithDefault(__x, Curry._2(DomainType.Model.Record.defaultWithId, param, id));
  };
  var idListToFilteredItems$1 = function (idList, normalized) {
    return removeOptionsFromList(Belt_List.map(idList, (function (param) {
                      return fromLocal(normalized, param);
                    })));
  };
  var reduceWithDefault = function (param) {
    return (function (param$1, param$2, param$3) {
        return reduce((function (__x) {
                      return Curry._2(DomainType.Model.Record.defaultWithId, param, __x);
                    }), fromLocal, DomainType.Action.reduce, param$1, param$2, param$3);
      });
  };
  var createReduceIdWithDefault = function (normalized, id, param, action) {
    return reduceWithDefault(param)(Promise.resolve(normalized), id, action);
  };
  var createPromiseReduceIdWithDefault = function (normalizedPromise, id, param, action) {
    return reduceWithDefault(param)(normalizedPromise, id, action);
  };
  var Local = {
    getRecord: fromLocal,
    getRecordWithDefault: getRecordWithDefault,
    idListToFilteredItems: idListToFilteredItems$1,
    reduceWithDefault: reduceWithDefault,
    createReduceIdWithDefault: createReduceIdWithDefault,
    createPromiseReduceIdWithDefault: createPromiseReduceIdWithDefault
  };
  var getRecordWithDefault$1 = function (normalized, id, param) {
    var __x = fromSchema(normalized, id);
    return Belt_Option.getWithDefault(__x, Curry._2(DomainType.Model.Record.defaultWithId, param, id));
  };
  var idListToFilteredItems$2 = function (idList, normalized) {
    return removeOptionsFromList(Belt_List.map(idList, (function (param) {
                      return fromSchema(normalized, param);
                    })));
  };
  var reduceWithDefault$1 = function (param) {
    return (function (param$1, param$2, param$3) {
        return reduce((function (__x) {
                      return Curry._2(DomainType.Model.Record.defaultWithId, param, __x);
                    }), fromSchema, DomainType.Action.reduce, param$1, param$2, param$3);
      });
  };
  var createReduceIdWithDefault$1 = function (normalized, id, param, action) {
    return reduceWithDefault$1(param)(Promise.resolve(normalized), id, action);
  };
  var createPromiseReduceIdWithDefault$1 = function (normalizedPromise, id, param, action) {
    return reduceWithDefault$1(param)(normalizedPromise, id, action);
  };
  var Remote = {
    getRecord: fromSchema,
    getRecordWithDefault: getRecordWithDefault$1,
    idListToFilteredItems: idListToFilteredItems$2,
    reduceWithDefault: reduceWithDefault$1,
    createReduceIdWithDefault: createReduceIdWithDefault$1,
    createPromiseReduceIdWithDefault: createPromiseReduceIdWithDefault$1
  };
  var AddSourceFunctions = function (ResourceReducer) {
    var updateWithPromiseDefault = function (normalized, updateNormalized, param, idType, action) {
      return Curry._1(updateNormalized, Curry._4(ResourceReducer.reduceWithDefault, param, normalized, idType, action));
    };
    var updateWithDefault = function (normalized, updateNormalized, param, idType, action) {
      return Curry._1(updateNormalized, Curry._4(ResourceReducer.reduceWithDefault, param, Promise.resolve(normalized), idType, action));
    };
    var createUpdateIdWithDefault = function (normalized, updateNormalized, id, param, action) {
      var actionFunc = Curry._3(ResourceReducer.createReduceIdWithDefault, normalized, id, param);
      return Curry._1(updateNormalized, Curry._1(actionFunc, action));
    };
    var createPromiseUpdateIdWithDefault = function (normalizedPromise, updateNormalized, id, param, action) {
      var actionFunc = Curry._3(ResourceReducer.createPromiseReduceIdWithDefault, normalizedPromise, id, param);
      return Curry._1(updateNormalized, Curry._1(actionFunc, action));
    };
    return {
            getRecord: ResourceReducer.getRecord,
            getRecordWithDefault: ResourceReducer.getRecordWithDefault,
            reduceWithDefault: ResourceReducer.reduceWithDefault,
            createReduceIdWithDefault: ResourceReducer.createReduceIdWithDefault,
            createPromiseReduceIdWithDefault: ResourceReducer.createPromiseReduceIdWithDefault,
            updateWithPromiseDefault: updateWithPromiseDefault,
            updateWithDefault: updateWithDefault,
            createUpdateIdWithDefault: createUpdateIdWithDefault,
            createPromiseUpdateIdWithDefault: createPromiseUpdateIdWithDefault
          };
  };
  var updateWithPromiseDefault = function (normalized, updateNormalized, param, idType, action) {
    return Curry._1(updateNormalized, reduceWithDefault(param)(normalized, idType, action));
  };
  var updateWithDefault = function (normalized, updateNormalized, param, idType, action) {
    return Curry._1(updateNormalized, reduceWithDefault(param)(Promise.resolve(normalized), idType, action));
  };
  var createUpdateIdWithDefault = function (normalized, updateNormalized, id, param, action) {
    var actionFunc = function (param$1) {
      return createReduceIdWithDefault(normalized, id, param, param$1);
    };
    return Curry._1(updateNormalized, actionFunc(action));
  };
  var createPromiseUpdateIdWithDefault = function (normalizedPromise, updateNormalized, id, param, action) {
    var actionFunc = function (param$1) {
      return createPromiseReduceIdWithDefault(normalizedPromise, id, param, param$1);
    };
    return Curry._1(updateNormalized, actionFunc(action));
  };
  var Local$1 = {
    getRecord: fromLocal,
    getRecordWithDefault: getRecordWithDefault,
    reduceWithDefault: reduceWithDefault,
    createReduceIdWithDefault: createReduceIdWithDefault,
    createPromiseReduceIdWithDefault: createPromiseReduceIdWithDefault,
    updateWithPromiseDefault: updateWithPromiseDefault,
    updateWithDefault: updateWithDefault,
    createUpdateIdWithDefault: createUpdateIdWithDefault,
    createPromiseUpdateIdWithDefault: createPromiseUpdateIdWithDefault
  };
  var updateWithPromiseDefault$1 = function (normalized, updateNormalized, param, idType, action) {
    return Curry._1(updateNormalized, reduceWithDefault$1(param)(normalized, idType, action));
  };
  var updateWithDefault$1 = function (normalized, updateNormalized, param, idType, action) {
    return Curry._1(updateNormalized, reduceWithDefault$1(param)(Promise.resolve(normalized), idType, action));
  };
  var createUpdateIdWithDefault$1 = function (normalized, updateNormalized, id, param, action) {
    var actionFunc = function (param$1) {
      return createReduceIdWithDefault$1(normalized, id, param, param$1);
    };
    return Curry._1(updateNormalized, actionFunc(action));
  };
  var createPromiseUpdateIdWithDefault$1 = function (normalizedPromise, updateNormalized, id, param, action) {
    var actionFunc = function (param$1) {
      return createPromiseReduceIdWithDefault$1(normalizedPromise, id, param, param$1);
    };
    return Curry._1(updateNormalized, actionFunc(action));
  };
  var Remote$1 = {
    getRecord: fromSchema,
    getRecordWithDefault: getRecordWithDefault$1,
    reduceWithDefault: reduceWithDefault$1,
    createReduceIdWithDefault: createReduceIdWithDefault$1,
    createPromiseReduceIdWithDefault: createPromiseReduceIdWithDefault$1,
    updateWithPromiseDefault: updateWithPromiseDefault$1,
    updateWithDefault: updateWithDefault$1,
    createUpdateIdWithDefault: createUpdateIdWithDefault$1,
    createPromiseUpdateIdWithDefault: createPromiseUpdateIdWithDefault$1
  };
  var Source = {
    Local: Local$1,
    Remote: Remote$1
  };
  var WithStore = function (NormalizeStore) {
    var AddStoreFunctions = function (ResourceReducer) {
      var getRecord = function (id) {
        return Curry._2(ResourceReducer.getRecord, Curry._1(NormalizeStore.getNormalized, /* () */0), id);
      };
      var getRecordWithDefault = function (id, param) {
        return Curry._3(ResourceReducer.getRecordWithDefault, Curry._1(NormalizeStore.getNormalized, /* () */0), id, param);
      };
      var reduceWithDefault = function (param) {
        return Curry._2(ResourceReducer.reduceWithDefault, param, Promise.resolve(Curry._1(NormalizeStore.getNormalized, /* () */0)));
      };
      var createReduceIdWithDefault = function (id, param) {
        return Curry._3(ResourceReducer.createReduceIdWithDefault, Curry._1(NormalizeStore.getNormalized, /* () */0), id, param);
      };
      var updateWithDefault = function (param, idType, action) {
        var updateNormalized = Curry._1(NormalizeStore.getUpdateNormalized, /* () */0);
        return Curry._1(updateNormalized, Curry._4(ResourceReducer.reduceWithDefault, param, Promise.resolve(Curry._1(NormalizeStore.getNormalized, /* () */0)), idType, action));
      };
      var createUpdateIdWithDefault = function (id, param) {
        var updateNormalized = Curry._1(NormalizeStore.getUpdateNormalized, /* () */0);
        return (function (action) {
            var actionFunc = Curry._3(ResourceReducer.createReduceIdWithDefault, Curry._1(NormalizeStore.getNormalized, /* () */0), id, param);
            return Curry._1(updateNormalized, Curry._1(actionFunc, action));
          });
      };
      return {
              getRecord: getRecord,
              getRecordWithDefault: getRecordWithDefault,
              reduceWithDefault: reduceWithDefault,
              createReduceIdWithDefault: createReduceIdWithDefault,
              updateWithDefault: updateWithDefault,
              createUpdateIdWithDefault: createUpdateIdWithDefault
            };
    };
    var getRecord = function (id) {
      return fromLocal(Curry._1(NormalizeStore.getNormalized, /* () */0), id);
    };
    var getRecordWithDefault$2 = function (id, param) {
      return getRecordWithDefault(Curry._1(NormalizeStore.getNormalized, /* () */0), id, param);
    };
    var reduceWithDefault$2 = function (param) {
      var partial_arg = Promise.resolve(Curry._1(NormalizeStore.getNormalized, /* () */0));
      var partial_arg$1 = reduceWithDefault(param);
      return (function (param, param$1) {
          return partial_arg$1(partial_arg, param, param$1);
        });
    };
    var createReduceIdWithDefault$2 = function (id, param) {
      var partial_arg = Curry._1(NormalizeStore.getNormalized, /* () */0);
      return (function (param$1) {
          return createReduceIdWithDefault(partial_arg, id, param, param$1);
        });
    };
    var updateWithDefault = function (param, idType, action) {
      var updateNormalized = Curry._1(NormalizeStore.getUpdateNormalized, /* () */0);
      return Curry._1(updateNormalized, reduceWithDefault(param)(Promise.resolve(Curry._1(NormalizeStore.getNormalized, /* () */0)), idType, action));
    };
    var createUpdateIdWithDefault = function (id, param) {
      var updateNormalized = Curry._1(NormalizeStore.getUpdateNormalized, /* () */0);
      return (function (action) {
          var partial_arg = Curry._1(NormalizeStore.getNormalized, /* () */0);
          var actionFunc = function (param$1) {
            return createReduceIdWithDefault(partial_arg, id, param, param$1);
          };
          return Curry._1(updateNormalized, Curry._1(actionFunc, action));
        });
    };
    var Local = {
      getRecord: getRecord,
      getRecordWithDefault: getRecordWithDefault$2,
      reduceWithDefault: reduceWithDefault$2,
      createReduceIdWithDefault: createReduceIdWithDefault$2,
      updateWithDefault: updateWithDefault,
      createUpdateIdWithDefault: createUpdateIdWithDefault
    };
    var getRecord$1 = function (id) {
      return fromSchema(Curry._1(NormalizeStore.getNormalized, /* () */0), id);
    };
    var getRecordWithDefault$3 = function (id, param) {
      return getRecordWithDefault$1(Curry._1(NormalizeStore.getNormalized, /* () */0), id, param);
    };
    var reduceWithDefault$3 = function (param) {
      var partial_arg = Promise.resolve(Curry._1(NormalizeStore.getNormalized, /* () */0));
      var partial_arg$1 = reduceWithDefault$1(param);
      return (function (param, param$1) {
          return partial_arg$1(partial_arg, param, param$1);
        });
    };
    var createReduceIdWithDefault$3 = function (id, param) {
      var partial_arg = Curry._1(NormalizeStore.getNormalized, /* () */0);
      return (function (param$1) {
          return createReduceIdWithDefault$1(partial_arg, id, param, param$1);
        });
    };
    var updateWithDefault$1 = function (param, idType, action) {
      var updateNormalized = Curry._1(NormalizeStore.getUpdateNormalized, /* () */0);
      return Curry._1(updateNormalized, reduceWithDefault$1(param)(Promise.resolve(Curry._1(NormalizeStore.getNormalized, /* () */0)), idType, action));
    };
    var createUpdateIdWithDefault$1 = function (id, param) {
      var updateNormalized = Curry._1(NormalizeStore.getUpdateNormalized, /* () */0);
      return (function (action) {
          var partial_arg = Curry._1(NormalizeStore.getNormalized, /* () */0);
          var actionFunc = function (param$1) {
            return createReduceIdWithDefault$1(partial_arg, id, param, param$1);
          };
          return Curry._1(updateNormalized, Curry._1(actionFunc, action));
        });
    };
    var Remote = {
      getRecord: getRecord$1,
      getRecordWithDefault: getRecordWithDefault$3,
      reduceWithDefault: reduceWithDefault$3,
      createReduceIdWithDefault: createReduceIdWithDefault$3,
      updateWithDefault: updateWithDefault$1,
      createUpdateIdWithDefault: createUpdateIdWithDefault$1
    };
    return {
            AddStoreFunctions: AddStoreFunctions,
            Local: Local,
            Remote: Remote
          };
  };
  return {
          Container: Container,
          Wrapper: Wrapper,
          normalizerGetItemFromSchema: normalizerGetItemFromSchema,
          idListToFilteredItems: idListToFilteredItems,
          reduce: reduce,
          fromLocal: fromLocal,
          fromSchema: fromSchema,
          GenerateConverterFunc: GenerateConverterFunc,
          LocalSourceContainer: LocalSourceContainer,
          RemoteSourceContainer: RemoteSourceContainer,
          Local: Local,
          Remote: Remote,
          AddSourceFunctions: AddSourceFunctions,
          Source: Source,
          WithStore: WithStore
        };
}

exports.tryAgainifNullOption = tryAgainifNullOption;
exports.removeOptionsFromList = removeOptionsFromList;
exports.DomainTypeConverter = DomainTypeConverter;
/* No side effect */
